import numpy as np
import numpy.testing as npt
import pytest
from pyxrf.gui_support.gpc_class import GlobalProcessingClasses

# =====================================================================
#                 class GlobalProcessingClasses (SecondaryWindow)


# fmt: off
@pytest.mark.parametrize("eline_keys", [
    [],
    ["Ca_K"],
    ["K_K", "Ca_K", "Fe_K", "positions", "some_key"]
])
# fmt: on
def test_gpc_get_maps_info_table_1(eline_keys):

    gpc = GlobalProcessingClasses()
    gpc.initialize()

    def _get_rand(v_low, v_high):
        return v_low + np.random.rand() * (v_high - v_low)

    # Generate tables
    range_table, limit_table, limit_table_norm, show_table = [], [], [], []
    for key in eline_keys:
        rng_min = _get_rand(0, 10)
        rng_max = _get_rand(20, 30)
        sel_min = _get_rand(rng_min, rng_min + (rng_max - rng_min) / 3)
        sel_max = _get_rand(rng_min + 2 * (rng_max - rng_min) / 3, rng_max)
        sel_min_norm = (sel_min - rng_min) / (rng_max - rng_min) * 100.0
        sel_max_norm = (sel_max - rng_min) / (rng_max - rng_min) * 100.0
        range_table.append([key, rng_min, rng_max])
        limit_table.append([key, sel_min, sel_max])
        limit_table_norm.append([key, sel_min_norm, sel_max_norm])
        show_table.append([key, np.random.rand() > 0.5])

    # 'Artificially' setup some fileds of 'gpc.img_model_adv'
    gpc.img_model_adv.map_keys = eline_keys
    gpc.img_model_adv.range_dict = {}
    gpc.img_model_adv.limit_dict = {}
    gpc.img_model_adv.stat_dict = {}
    for n, key in enumerate(eline_keys):
        gpc.img_model_adv.range_dict[key] = {}
        gpc.img_model_adv.range_dict[key]["low"] = range_table[n][1]
        gpc.img_model_adv.range_dict[key]["high"] = range_table[n][2]
        gpc.img_model_adv.limit_dict[key] = {}
        gpc.img_model_adv.limit_dict[key]["low"] = limit_table_norm[n][1]
        gpc.img_model_adv.limit_dict[key]["high"] = limit_table_norm[n][2]
        gpc.img_model_adv.stat_dict[key] = show_table[n][1]

    # Now run the test
    rng_table, lim_table, sh_table = gpc.get_maps_info_table()
    assert rng_table == range_table, "Returned table of ranges does not match the expected."
    assert [_[0] for _ in lim_table] == [
        _[0] for _ in limit_table
    ], "Keys in the returned table of limits do not match the expected."
    npt.assert_array_almost_equal(
        [_[1:] for _ in lim_table],
        [_[1:] for _ in limit_table],
        err_msg="Returned limits don't match the expected",
    )
    assert sh_table == show_table, "'show' status table don't match the expected table"


def test_gpc_get_maps_info_table_2():
    """Failing cases for `GlobalProcessingClasses.get_maps_info_table()`"""

    gpc = GlobalProcessingClasses()
    gpc.initialize()

    eline_keys = ["abc", "def"]  # Some random keys
    gpc.img_model_adv.map_keys = eline_keys
    correct_dict = {_: {"low": 0.0, "high": 1.0} for _ in eline_keys}
    correct_stat_dict = {_: True for _ in eline_keys}

    # Modified 'limit_dict'
    modified_dict = correct_dict.copy()
    modified_dict["extra_key"] = {"low": 0.0, "high": 1.0}

    modified_stat_dict = correct_stat_dict.copy()
    modified_stat_dict["extra_key"] = False

    gpc.img_model_adv.range_dict = correct_dict
    gpc.img_model_adv.limit_dict = modified_dict
    gpc.img_model_adv.stat_dict = correct_stat_dict

    with pytest.raises(RuntimeError, match="list of keys in 'limit_dict'"):
        gpc.get_maps_info_table()

    gpc.img_model_adv.range_dict = modified_dict
    gpc.img_model_adv.limit_dict = correct_dict
    gpc.img_model_adv.stat_dict = correct_stat_dict

    with pytest.raises(RuntimeError, match="list of keys in 'range_dict'"):
        gpc.get_maps_info_table()

    gpc.img_model_adv.range_dict = correct_dict
    gpc.img_model_adv.limit_dict = correct_dict
    gpc.img_model_adv.stat_dict = modified_stat_dict

    with pytest.raises(RuntimeError, match="list of keys in 'stat_dict'"):
        gpc.get_maps_info_table()



from qtpy.QtWidgets import (
    QHBoxLayout,
    QVBoxLayout,
    QLabel,
    QComboBox,
    QMessageBox,
    QTableWidget,
    QTableWidgetItem,
    QHeaderView,
    QPushButton,
)
from qtpy.QtGui import QBrush, QColor, QPalette
from qtpy.QtCore import Qt, Signal, Slot, QThreadPool, QRunnable

from .useful_widgets import get_background_css, set_tooltip, ComboBoxNamedNoWheel, SecondaryWindow

import logging

logger = logging.getLogger(__name__)


fitting_preset_names = {
    "None": "None",
    "fit_with_tail": "With Tail",
    "free_more": "Free",
    "e_calibration": "E-axis",
    "linear": "Area",
    "adjust_element1": "Custom 1",
    "adjust_element2": "Custom 2",
    "adjust_element3": "Custom 3",
}

value_names = {
    "value": "Value",
    "min": "Min",
    "max": "Max",
}


class WndDetailedFittingParams(SecondaryWindow):

    # Signal that is sent (to main window) to update global state of the program
    update_global_state = Signal()
    computations_complete = Signal(object)

    def __init__(self, *, window_title, gpc, gui_vars):
        super().__init__()

        # Global processing classes
        self.gpc = gpc
        # Global GUI variables (used for control of GUI state)
        self.gui_vars = gui_vars

        # Reference to the main window. The main window will hold
        #   references to all non-modal windows that could be opened
        #   from multiple places in the program.
        self.ref_main_window = self.gui_vars["ref_main_window"]

        self.update_global_state.connect(self.ref_main_window.update_widget_state)

        self._enable_events = False

        self._dialog_data = {}

        self._load_dialog_data()
        self._selected_index = 0
        self._selected_eline = "-"

        self.setWindowTitle(window_title)
        self.setMinimumWidth(1100)
        self.setMinimumHeight(500)
        self.resize(1100, 500)

        hbox_el_select = self._setup_element_selection()
        self._setup_table()

        vbox = QVBoxLayout()
        vbox.addLayout(hbox_el_select)
        vbox.addWidget(self.table)

        self.setLayout(vbox)

        self.update_form_data()
        self._enable_events = True
        self._data_changed = False

    def _setup_element_selection(self):

        self.combo_element_sel = QComboBox()
        self.combo_element_sel.setMinimumWidth(200)
        self.combo_element_sel.currentIndexChanged.connect(self.combo_element_sel_current_index_changed)

        self.pb_apply = QPushButton("Apply")
        self.pb_apply.setEnabled(False)
        self.pb_apply.clicked.connect(self.pb_apply_clicked)
        self.pb_cancel = QPushButton("Cancel")
        self.pb_cancel.setEnabled(False)
        self.pb_cancel.clicked.connect(self.pb_cancel_clicked)

        hbox = QHBoxLayout()
        hbox.addWidget(QLabel("Select element:"))
        hbox.addWidget(self.combo_element_sel)
        hbox.addStretch(1)
        hbox.addWidget(self.pb_apply)
        hbox.addWidget(self.pb_cancel)

        return hbox

    def _setup_table(self):

        self._value_keys = ("value", "min", "max")

        # Labels for horizontal header
        labels_presets = [fitting_preset_names[_] for _ in self._fit_strategy_list]
        labels_values = [value_names[_] for _ in self._value_keys]
        self.tbl_labels = ["Name", "E, keV"] + labels_values + labels_presets
        # Labels for editable columns
        self.tbl_cols_editable = ("Value", "Min", "Max")
        # Labels for the columns that contain combo boxes
        self.tbl_cols_combobox = labels_presets
        # The list of columns with fixed size
        self.tbl_cols_stretch = ("Value", "Min", "Max")
        # Table item representation if different from default
        self.tbl_format = {"E, keV": ".4f", "Value": ".8g", "Min": ".8g", "Max": ".8g"}

        # Combobox items. All comboboxes in the table contain identical list of items.
        self.combo_items = self._bound_options

        self._combo_list = []

        self.table = QTableWidget()
        self.table.setColumnCount(len(self.tbl_labels))
        self.table.verticalHeader().hide()
        self.table.setHorizontalHeaderLabels(self.tbl_labels)

        self.table.setStyleSheet("QTableWidget::item{color: black;}")

        header = self.table.horizontalHeader()
        for n, lbl in enumerate(self.tbl_labels):
            # Set stretching for the columns
            if lbl in self.tbl_cols_stretch:
                header.setSectionResizeMode(n, QHeaderView.Stretch)
            else:
                header.setSectionResizeMode(n, QHeaderView.ResizeToContents)

        self.table.itemChanged.connect(self.tbl_elines_item_changed)

    def _fill_table(self, table_contents):
        self._enable_events = False

        # Clear the list of combo boxes
        for item in self._combo_list:
            item.currentIndexChanged.disconnect(self.combo_strategy_current_index_changed)
        self._combo_list = []

        self.table.clearContents()

        self.table.setRowCount(len(table_contents))
        for nr, row in enumerate(table_contents):
            n_fit_strategy = 0
            row_name = row[0]
            for nc, entry in enumerate(row):
                label = self.tbl_labels[nc]

                # Set alternating background colors for the table rows
                #   Make background for editable items a little brighter
                brightness = 240 if label in self.tbl_cols_editable else 220
                if nr % 2:
                    rgb_bckg = (255, brightness, brightness)
                else:
                    rgb_bckg = (brightness, 255, brightness)

                if label not in self.tbl_cols_combobox:
                    if label in self.tbl_format and not isinstance(entry, str):
                        fmt = self.tbl_format[self.tbl_labels[nc]]
                        s = ("{:" + fmt + "}").format(entry)
                    else:
                        s = f"{entry}"

                    item = QTableWidgetItem(s)
                    if nc > 0:
                        item.setTextAlignment(Qt.AlignRight | Qt.AlignVCenter)

                    # Set all columns not editable (unless needed)
                    if label not in self.tbl_cols_editable:
                        item.setFlags(item.flags() & ~Qt.ItemIsEditable)

                    # Make all items not selectable (we are not using selections)
                    item.setFlags(item.flags() & ~Qt.ItemIsSelectable)

                    # Note, that there is no way to set style sheet for QTableWidgetItem
                    item.setBackground(QBrush(QColor(*rgb_bckg)))

                    self.table.setItem(nr, nc, item)

                else:
                    if n_fit_strategy < len(self._fit_strategy_list):
                        combo_name = f"{row_name},{self._fit_strategy_list[n_fit_strategy]}"
                    else:
                        combo_name = ""
                    n_fit_strategy += 1

                    item = ComboBoxNamedNoWheel(name=combo_name)

                    # Set text color for QComboBox widget (necessary if the program is used with Dark theme)
                    pal = item.palette()
                    pal.setColor(QPalette.ButtonText, Qt.black)
                    item.setPalette(pal)
                    # Set text color for drop-down view (necessary if the program is used with Dark theme)
                    pal = item.view().palette()
                    pal.setColor(QPalette.Text, Qt.black)
                    item.view().setPalette(pal)

                    css1 = get_background_css(rgb_bckg, widget="QComboBox", editable=False)
                    css2 = get_background_css(rgb_bckg, widget="QWidget", editable=False)
                    item.setStyleSheet(css2 + css1)

                    item.addItems(self.combo_items)
                    if item.findText(entry) < 0:
                        logger.warning(f"Text '{entry}' is not found. The ComboBox is not set properly.")
                    item.setCurrentText(entry)  # Try selecting the item anyway
                    self.table.setCellWidget(nr, nc, item)

                    item.currentIndexChanged.connect(self.combo_strategy_current_index_changed)
                    self._combo_list.append(item)

        self._enable_events = True

    def _set_tooltips(self):
        set_tooltip(self.pb_apply, "Save changes and <b>update plots</b>.")
        set_tooltip(self.pb_cancel, "<b>Discard</b> all changes.")
        set_tooltip(
            self.combo_element_sel,
            "Select K, L or M <b>emission line</b> to edit the optimization parameters "
            "used for the line during total spectrum fitting.",
        )
        set_tooltip(
            self.table,
            "Edit optimization parameters for the selected emission line. "
            "Processing presets may be configured by specifying optimization strategy "
            "for each parameter may be selected. A preset for each fitting step "
            "of the total spectrum fitting may be selected in <b>Model</b> tab.",
        )

    def combo_element_sel_current_index_changed(self, index):
        self._selected_index = index
        try:
            self._selected_eline = self._eline_list[index]
        except Exception:
            self._selected_eline = "-"
        self._update_table()

    def combo_strategy_current_index_changed(self, name, index):
        if self._enable_events:
            try:
                name_row, name_strategy = name.split(",")
                option = self._bound_options[index]
                self._param_dict[name_row][name_strategy] = option
            except Exception as ex:
                logger.error(f"Error occurred while changing strategy options: {ex}")

            self._data_changed = True
            self._validate_all()

    def tbl_elines_item_changed(self, item):
        if self._enable_events:
            try:
                n_row, n_col = self.table.row(item), self.table.column(item)
                n_key = n_col - 2
                if n_key < 0 or n_key >= len(self._value_keys):
                    raise RuntimeError(f"Incorrect column {n_col}")
                value_key = self._value_keys[n_key]
                eline_key = self._table_contents[n_row][0]
                try:
                    value = float(item.text())
                    self._param_dict[eline_key][value_key] = value
                except Exception:
                    value = self._param_dict[eline_key][value_key]
                    item.setText(f"{value:.8g}")

                self._data_changed = True
                self._validate_all()

            except Exception as ex:
                logger.error(f"Error occurred while setting edited value: {ex}")

    def pb_apply_clicked(self):
        """Save dialog data and update plots"""
        self.save_form_data()

    def pb_cancel_clicked(self):
        """Reload data (discard all changes)"""
        self.update_form_data()

    def _set_combo_element_sel_items(self):
        element_list = self._eline_list
        self.combo_element_sel.clear()
        self.combo_element_sel.addItems(element_list)
        # Deselect all (this should clear the table)
        self.select_eline(self._selected_eline)

    def _set_dialog_data(self, dialog_data):
        self._param_dict = dialog_data["param_dict"]
        self._eline_list = dialog_data["eline_list"]
        self._eline_key_dict = dialog_data["eline_key_dict"]
        self._eline_energy_dict = dialog_data["eline_energy_dict"]
        self._other_param_list = dialog_data["other_param_list"]
        self._fit_strategy_list = dialog_data["fit_strategy_list"]
        self._bound_options = dialog_data["bound_options"]

    def _show_all(self):
        selected_eline = self._selected_eline
        self._set_combo_element_sel_items()
        self.select_eline(selected_eline)
        self._update_table()

    def _update_table(self):
        self._enable_events = False

        eline_list = []
        if "shared" in self._selected_eline.lower():  # Shared parameters
            eline_list = self._other_param_list
            energy_list = [""] * len(eline_list)
        elif self._selected_eline == self._eline_list[self._selected_index]:  # Emission lines
            eline = self._selected_eline
            eline_list = self._eline_key_dict[eline]
            energy_list = self._eline_energy_dict[eline]

        self._table_contents = []
        for n, key in enumerate(eline_list):
            data = [
                key,
                energy_list[n],
                self._param_dict[key]["value"],
                self._param_dict[key]["min"],
                self._param_dict[key]["max"],
            ]
            for strategy in self._fit_strategy_list:
                data.append(self._param_dict[key][strategy])
            self._table_contents.append(data)

        self._fill_table(self._table_contents)
        self._enable_events = True

    def select_eline(self, eline):
        if eline in self._eline_list:
            index = self._eline_list.index(eline)
        elif self._eline_list:
            index = 0
            eline = self._eline_list[0]
        else:
            index = -1
            eline = "-"
        self._selected_eline = eline
        self._selected_index = index
        self.combo_element_sel.setCurrentIndex(index)
        self._update_table()

    def update_widget_state(self, condition=None):
        # Update the state of the menu bar
        state = not self.gui_vars["gui_state"]["running_computations"]
        self.setEnabled(state)

        if condition == "tooltips":
            self._set_tooltips()

    def _validate_all(self):
        self.pb_apply.setEnabled(self._data_changed)
        self.pb_cancel.setEnabled(self._data_changed)

    def _load_dialog_data(self):
        ...

    def _save_dialog_data_function(self):
        ...

    def update_form_data(self):
        self._load_dialog_data()
        self._show_all()
        self._data_changed = False
        self._validate_all()

    def save_form_data(self):
        if self._data_changed:
            f_save_data = self._save_dialog_data_function()

            def cb(dialog_data):
                try:
                    f_save_data(dialog_data)
                    success, msg = True, ""
                except Exception as ex:
                    success, msg = False, str(ex)
                return {"success": success, "msg": msg}

            self._compute_in_background(cb, self.slot_save_form_data, dialog_data=self._dialog_data)

    @Slot(object)
    def slot_save_form_data(self, result):
        self._recover_after_compute(self.slot_save_form_data)

        if not result["success"]:
            msg = result["msg"]
            msgbox = QMessageBox(
                QMessageBox.Critical, "Failed to Apply Fit Parameters", msg, QMessageBox.Ok, parent=self
            )
            msgbox.exec()
        else:
            self._data_changed = False
            self._validate_all()

        self.gui_vars["gui_state"]["state_model_fit_exists"] = False
        self.update_global_state.emit()

    def _compute_in_background(self, func, slot, *args, **kwargs):
        """
        Run function `func` in a background thread. Send the signal
        `self.computations_complete` once computation is finished.

        Parameters
        ----------
        func: function
            Reference to a function that is supposed to be executed at the background.
            The function return value is passed as a signal parameter once computation is
            complete.
        slot: qtpy.QtCore.Slot or None
            Reference to a slot. If not None, then the signal `self.computation_complete`
            is connected to this slot.
        args, kwargs
            arguments of the function `func`.
        """
        signal_complete = self.computations_complete

        def func_to_run(func, *args, **kwargs):
            class RunTask(QRunnable):
                def run(self):
                    result_dict = func(*args, **kwargs)
                    signal_complete.emit(result_dict)

            return RunTask()

        if slot is not None:
            self.computations_complete.connect(slot)
        self.gui_vars["gui_state"]["running_computations"] = True
        self.update_global_state.emit()
        QThreadPool.globalInstance().start(func_to_run(func, *args, **kwargs))

    def _recover_after_compute(self, slot):
        """
        The function should be called after the signal `self.computations_complete` is
        received. The slot should be the same as the one used when calling
        `self.compute_in_background`.
        """
        if slot is not None:
            self.computations_complete.disconnect(slot)
        self.gui_vars["gui_state"]["running_computations"] = False
        self.update_global_state.emit()


class WndDetailedFittingParamsLines(WndDetailedFittingParams):
    def __init__(self, *, gpc, gui_vars):
        window_title = "Fitting Parameters for Individual Emission Lines"
        super().__init__(window_title=window_title, gpc=gpc, gui_vars=gui_vars)

    def _load_dialog_data(self):
        self._dialog_data = self.gpc.get_detailed_fitting_params_lines()
        self._set_dialog_data(self._dialog_data)

    def _save_dialog_data_function(self):
        def f(dialog_data):
            self.gpc.set_detailed_fitting_params(dialog_data)

        return f


class WndDetailedFittingParamsShared(WndDetailedFittingParams):
    def __init__(self, *, gpc, gui_vars):
        window_title = "Shared Detailed Fitting Parameters"
        super().__init__(window_title=window_title, gpc=gpc, gui_vars=gui_vars)

    def _load_dialog_data(self):
        self._dialog_data = self.gpc.get_detailed_fitting_params_shared()
        self._set_dialog_data(self._dialog_data)

    def _save_dialog_data_function(self):
        def f(dialog_data):
            self.gpc.set_detailed_fitting_params(dialog_data)

        return f



from qtpy.QtWidgets import QVBoxLayout, QDialog, QDialogButtonBox, QTextEdit

import logging

logger = logging.getLogger(__name__)


class DialogViewMetadata(QDialog):
    def __init__(self):

        super().__init__()

        self.resize(500, 500)
        self.setWindowTitle("Run Metadata")

        self.te_meta = QTextEdit()
        self.te_meta.setReadOnly(True)

        # 'Close' button box
        button_box = QDialogButtonBox(QDialogButtonBox.Close)
        button_box.accepted.connect(self.accept)
        button_box.rejected.connect(self.reject)

        vbox = QVBoxLayout()
        vbox.addWidget(self.te_meta)
        vbox.addWidget(button_box)
        self.setLayout(vbox)

    def setText(self, text):
        self.te_meta.setText(text)



from qtpy.QtWidgets import QWidget, QVBoxLayout, QHBoxLayout, QComboBox, QCheckBox, QSpacerItem
from qtpy.QtCore import Signal, Slot

from .useful_widgets import set_tooltip, global_gui_variables

from matplotlib.backends.backend_qt5agg import (
    FigureCanvasQTAgg as FigureCanvas,
    NavigationToolbar2QT as NavigationToolbar,
)

from .wd_rgb_selection import RgbSelectionWidget

import logging

logger = logging.getLogger(__name__)


class PlotRgbMaps(QWidget):

    signal_rgb_maps_dataset_selection_changed = Signal()
    signal_rgb_maps_norm_changed = Signal()
    signal_redraw_rgb_maps = Signal()

    def __init__(self, *, gpc, gui_vars):
        super().__init__()

        self._enable_plot_updates = True
        self._changes_exist = False
        self._enable_events = False

        # Global processing classes
        self.gpc = gpc
        # Global GUI variables (used for control of GUI state)
        self.gui_vars = gui_vars

        self.combo_select_dataset = QComboBox()
        self.combo_select_dataset.setSizeAdjustPolicy(QComboBox.AdjustToContents)

        self.combo_normalization = QComboBox()

        self.cb_interpolate = QCheckBox("Interpolate")
        self.cb_interpolate.setChecked(self.gpc.get_rgb_maps_grid_interpolate())
        self.cb_interpolate.toggled.connect(self.cb_interpolate_toggled)

        self.cb_quantitative = QCheckBox("Quantitative")
        self.cb_quantitative.setChecked(self.gpc.get_maps_quant_norm_enabled())
        self.cb_quantitative.toggled.connect(self.cb_quantitative_toggled)

        self.combo_pixels_positions = QComboBox()
        self._pix_pos_values = ["Pixels", "Positions"]
        self.combo_pixels_positions.addItems(self._pix_pos_values)
        self.combo_pixels_positions.setCurrentIndex(self._pix_pos_values.index(self.gpc.get_maps_pixel_or_pos()))
        self.combo_pixels_positions.currentIndexChanged.connect(self.combo_pixels_positions_current_index_changed)

        self.mpl_canvas = FigureCanvas(self.gpc.img_model_rgb.fig)
        self.mpl_toolbar = NavigationToolbar(self.mpl_canvas, self)

        # Keep layout without change when canvas is hidden (invisible)
        sp_retain = self.mpl_canvas.sizePolicy()
        sp_retain.setRetainSizeWhenHidden(True)
        self.mpl_canvas.setSizePolicy(sp_retain)

        self.rgb_selection = RgbSelectionWidget()
        self.slot_update_dataset_info()

        self.rgb_selection.signal_update_map_selections.connect(self._update_map_selections)

        self.widgets_enable_events(True)

        vbox = QVBoxLayout()
        hbox = QHBoxLayout()
        hbox.addWidget(self.combo_select_dataset)
        hbox.addWidget(self.combo_normalization)
        hbox.addStretch(1)
        hbox.addWidget(self.cb_quantitative)
        hbox.addWidget(self.cb_interpolate)
        hbox.addWidget(self.combo_pixels_positions)
        vbox.addLayout(hbox)
        vbox.addWidget(self.mpl_toolbar)
        vbox.addWidget(self.mpl_canvas)
        hbox = QHBoxLayout()
        hbox.addSpacerItem(QSpacerItem(0, 0))
        hbox.addWidget(self.rgb_selection)
        hbox.addSpacerItem(QSpacerItem(0, 0))
        vbox.addLayout(hbox)
        self.setLayout(vbox)

        self._set_tooltips()

    def _set_tooltips(self):
        set_tooltip(self.combo_select_dataset, "Select <b>dataset</b>.")
        set_tooltip(self.combo_normalization, "Select <b>scaler</b> for normalization of displayed XRF maps.")
        set_tooltip(self.cb_interpolate, "Interpolate coordinates to <b>uniform grid</b>.")
        set_tooltip(
            self.cb_quantitative,
            "Normalize the displayed XRF maps using loaded <b>Quantitative Calibration</b> data.",
        )
        set_tooltip(
            self.combo_pixels_positions, "Switch axes units between <b>pixels</b> and <b>positional units</b>."
        )
        set_tooltip(
            self.rgb_selection,
            "Select XRF Maps displayed in <b>Red</b>, <b>Green</b> and "
            "<b>Blue</b> colors and adjust the range of <b>intensity</b> for each "
            "displayed map.",
        )

    def widgets_enable_events(self, status):
        if status:
            if not self._enable_events:
                self.combo_select_dataset.currentIndexChanged.connect(
                    self.combo_select_dataset_current_index_changed
                )
                self.combo_normalization.currentIndexChanged.connect(
                    self.combo_normalization_current_index_changed
                )
                self._enable_events = True
        else:
            if self._enable_events:
                self.combo_select_dataset.currentIndexChanged.disconnect(
                    self.combo_select_dataset_current_index_changed
                )
                self.combo_normalization.currentIndexChanged.disconnect(
                    self.combo_normalization_current_index_changed
                )
                self._enable_events = False

    def update_widget_state(self, condition=None):
        if condition == "tooltips":
            self._set_tooltips()
        self.mpl_toolbar.setVisible(self.gui_vars["show_matplotlib_toolbar"])

        # Hide Matplotlib canvas during computations
        state_compute = global_gui_variables["gui_state"]["running_computations"]
        self.mpl_canvas.setVisible(not state_compute)

    def combo_select_dataset_current_index_changed(self, index):
        self.gpc.set_rgb_maps_selected_dataset(index + 1)
        self._update_dataset()
        self.signal_rgb_maps_dataset_selection_changed.emit()

    @Slot()
    def combo_select_dataset_update_current_index(self):
        index = self.gpc.get_rgb_maps_selected_dataset()
        self.combo_select_dataset.setCurrentIndex(index - 1)

    def combo_normalization_current_index_changed(self, index):
        self.gpc.set_rgb_maps_scaler_index(index)
        self.slot_update_ranges()
        self.signal_rgb_maps_norm_changed.emit()

    def combo_pixels_positions_current_index_changed(self, index):
        self.gpc.set_rgb_maps_pixel_or_pos(self._pix_pos_values[index])

    def cb_interpolate_toggled(self, state):
        self.gpc.set_rgb_maps_grid_interpolate(state)

    def cb_quantitative_toggled(self, state):
        self.gpc.set_rgb_maps_quant_norm_enabled(state)
        self.slot_update_ranges()
        self.signal_rgb_maps_norm_changed.emit()

    @Slot()
    def slot_update_dataset_info(self):
        self._update_dataset_list()
        self._update_dataset()
        self.cb_quantitative.setChecked(self.gpc.get_rgb_maps_quant_norm_enabled())

    def _update_dataset(self):
        self._update_scalers()

        # Update ranges in the RGB selection widget
        range_table, limit_table, rgb_dict = self.gpc.get_rgb_maps_info_table()
        self.rgb_selection.set_ranges_and_limits(
            range_table=range_table, limit_table=limit_table, rgb_dict=rgb_dict
        )

    @Slot()
    def slot_update_ranges(self):
        """Update only ranges and selections for the emission lines"""
        range_table, limit_table, _ = self.gpc.get_rgb_maps_info_table()
        self.rgb_selection.set_ranges_and_limits(range_table=range_table, limit_table=limit_table)

    def _update_dataset_list(self):
        self.widgets_enable_events(False)
        dataset_list, dset_sel = self.gpc.get_rgb_maps_dataset_list()
        self._dataset_list = dataset_list.copy()
        self.combo_select_dataset.clear()
        self.combo_select_dataset.addItems(self._dataset_list)
        # No item should be selected if 'dset_sel' is 0
        self.combo_select_dataset.setCurrentIndex(dset_sel - 1)
        self.widgets_enable_events(True)

    def _update_scalers(self):
        self.widgets_enable_events(False)
        scalers, scaler_sel = self.gpc.get_rgb_maps_scaler_list()
        self._scaler_list = ["Normalize by ..."] + scalers
        self.combo_normalization.clear()
        self.combo_normalization.addItems(self._scaler_list)
        self.combo_normalization.setCurrentIndex(scaler_sel)
        self.widgets_enable_events(True)

    @Slot()
    def _update_map_selections(self):
        """Upload the selections (limit table) and update plot"""
        if self._enable_plot_updates:
            self._changes_exist = False
            self.gpc.set_rgb_maps_limit_table(self.rgb_selection._limit_table, self.rgb_selection._rgb_dict)
            self._redraw_maps()

    def _redraw_maps(self):
        logger.debug("Redrawing XRF Maps")
        self.gpc.redraw_rgb_maps()
        self.signal_redraw_rgb_maps.emit()



param_data = {
    "coherent_sct_amplitude": {
        "adjust_element1": "fixed",
        "adjust_element2": "fixed",
        "adjust_element3": "fixed",
        "bound_type": "none",
        "e_calibration": "none",
        "fit_with_tail": "none",
        "free_more": "none",
        "linear": "none",
        "max": 1e9,
        "min": 1.0,
        "default": 100.0,
        "value": 91.84486613900303,
    },
    "coherent_sct_energy": {
        "adjust_element1": "fixed",
        "adjust_element2": "fixed",
        "adjust_element3": "fixed",
        "bound_type": "fixed",
        "description": "Incident E [keV]",
        "e_calibration": "fixed",
        "fit_with_tail": "fixed",
        "free_more": "lohi",
        "linear": "fixed",
        "max": 13.0,
        "min": 9.0,
        "default": 12.0,
        "value": 12.0,
    },
    "compton_amplitude": {
        "adjust_element1": "fixed",
        "adjust_element2": "fixed",
        "adjust_element3": "fixed",
        "bound_type": "none",
        "e_calibration": "none",
        "fit_with_tail": "none",
        "free_more": "none",
        "linear": "none",
        "max": 1e9,
        "min": 0.0,
        "default": 100.0,
        "value": 213.7603636952067,
    },
    "compton_angle": {
        "adjust_element1": "fixed",
        "adjust_element2": "fixed",
        "adjust_element3": "fixed",
        "bound_type": "fixed",
        "e_calibration": "fixed",
        "fit_with_tail": "lohi",
        "free_more": "lohi",
        "linear": "fixed",
        "max": 105.0,
        "min": 70.0,
        "default": 90.0,
        "value": 90.0,
    },
    "compton_f_step": {
        "adjust_element1": "fixed",
        "adjust_element2": "fixed",
        "adjust_element3": "fixed",
        "bound_type": "fixed",
        "e_calibration": "fixed",
        "fit_with_tail": "fixed",
        "free_more": "lohi",
        "linear": "fixed",
        "max": 0.01,
        "min": 0.0,
        "default": 0.005,
        "value": 0.00842597013124,
    },
    "compton_f_tail": {
        "adjust_element1": "fixed",
        "adjust_element2": "fixed",
        "adjust_element3": "fixed",
        "bound_type": "fixed",
        "e_calibration": "fixed",
        "fit_with_tail": "lohi",
        "free_more": "fixed",
        "linear": "fixed",
        "max": 4.0,
        "min": 1.5,
        "default": 2.5,
        "value": 2.5,
    },
    "compton_fwhm_corr": {
        "adjust_element1": "fixed",
        "adjust_element2": "fixed",
        "adjust_element3": "fixed",
        "bound_type": "fixed",
        "description": "fwhm Coef, Compton",
        "e_calibration": "fixed",
        "fit_with_tail": "lohi",
        "free_more": "lohi",
        "linear": "fixed",
        "max": 3.0,
        "min": 1.5,
        "default": 2.0,
        "value": 2.0,
    },
    "compton_gamma": {
        "adjust_element1": "fixed",
        "adjust_element2": "fixed",
        "adjust_element3": "fixed",
        "bound_type": "fixed",
        "e_calibration": "fixed",
        "fit_with_tail": "fixed",
        "free_more": "lohi",
        "linear": "fixed",
        "max": 4.2,
        "min": 3.8,
        "default": 2.5,
        "value": 4.161818621138601,
    },
    "compton_hi_f_tail": {
        "adjust_element1": "fixed",
        "adjust_element2": "fixed",
        "adjust_element3": "fixed",
        "bound_type": "fixed",
        "e_calibration": "fixed",
        "fit_with_tail": "fixed",
        "free_more": "fixed",
        "linear": "fixed",
        "max": 1.0,
        "min": 1e-06,
        "default": 0.0001,
        "value": 0.0001,
    },
    "compton_hi_gamma": {
        "adjust_element1": "fixed",
        "adjust_element2": "fixed",
        "adjust_element3": "fixed",
        "bound_type": "fixed",
        "e_calibration": "fixed",
        "fit_with_tail": "fixed",
        "free_more": "fixed",
        "linear": "fixed",
        "max": 3.0,
        "min": 0.025,
        "default": 0.028,
        "value": 0.027,
    },
    "e_linear": {
        "adjust_element1": "fixed",
        "adjust_element2": "fixed",
        "adjust_element3": "fixed",
        "bound_type": "lohi",
        "description": "E Calib. Coef, a1",
        "e_calibration": "lohi",
        "fit_with_tail": "fixed",
        "free_more": "lohi",
        "linear": "fixed",
        "max": 0.013,
        "min": 0.009,
        "tool_tip": "E(channel) = a0 + a1*channel+ a2*channel**2",
        "default": 0.01,
        "value": 0.01,
    },
    "e_offset": {
        "adjust_element1": "fixed",
        "adjust_element2": "fixed",
        "adjust_element3": "fixed",
        "bound_type": "lohi",
        "description": "E Calib. Coef, a0",
        "e_calibration": "lohi",
        "fit_with_tail": "fixed",
        "free_more": "hi",
        "linear": "fixed",
        "max": 0.2,
        "min": -0.2,
        "tool_tip": "E(channel) = a0 + a1*channel+ a2*channel**2",
        "default": 0.0,
        "value": 0.0,
    },
    "e_quadratic": {
        "adjust_element1": "fixed",
        "adjust_element2": "fixed",
        "adjust_element3": "fixed",
        "bound_type": "lohi",
        "description": "E Calib. Coef, a2",
        "e_calibration": "lohi",
        "fit_with_tail": "fixed",
        "free_more": "lohi",
        "linear": "fixed",
        "max": 1e-06,
        "min": -1e-06,
        "tool_tip": "E(channel) = a0 + a1*channel+ a2*channel**2",
        "default": 0.0,
        "value": 0.0,
    },
    "fwhm_fanoprime": {
        "adjust_element1": "fixed",
        "adjust_element2": "fixed",
        "adjust_element3": "fixed",
        "bound_type": "fixed",
        "description": "fwhm Coef, b2",
        "e_calibration": "fixed",
        "fit_with_tail": "lohi",
        "free_more": "lohi",
        "linear": "fixed",
        "max": 0.0002,
        "min": 1e-05,
        "tool_tip": (
            "width**2 = (b1/2.3548)**2 + 3.85*b2*E, 3.85keV is electron-hole pair creation energy in silicon"
        ),
        "default": 0.000114,
        "value": 0.000114,
    },
    "fwhm_offset": {
        "adjust_element1": "fixed",
        "adjust_element2": "fixed",
        "adjust_element3": "fixed",
        "bound_type": "fixed",
        "description": "fwhm Coef, b1 [keV]",
        "e_calibration": "fixed",
        "fit_with_tail": "lohi",
        "free_more": "lohi",
        "linear": "fixed",
        "max": 0.2,
        "min": 0.05,
        "tool_tip": "width**2 = (b1/2.3548)**2 + 3.85*b2*E",
        "default": 0.1,
        "value": 0.1,
    },
    "non_fitting_values": {
        "element_list": "Al_K, S_K, Cl_K, Ar_K, Ca_K, Cr_K, Fe_K, Co_K, Ni_K, Cu_K, Pm_L, Gd_L",
        "energy_bound_high": {"default_value": 12.8, "description": "E high [keV]", "value": 12.8},
        "energy_bound_low": {"default_value": 1.0, "description": "E low [keV]", "value": 1.0},
        "epsilon": 3.85,
        "escape_ratio": 0.005,
        "background_width": 0.5,
    },
}



import pytest
import re
import numpy as np
import numpy.testing as npt
from pyxrf.core.xrf_utils import (
    get_element_atomic_number,
    validate_element_str,
    parse_compound_formula,
    split_compound_mass,
    get_supported_eline_list,
    check_if_eline_supported,
    check_if_eline_is_activated,
    generate_eline_list,
)


# fmt: off
@pytest.mark.parametrize("element_number", [
    ("He", 2), ("H", 1), ("Fe", 26),
    # Failing cases, 0 is returned
    ("he", 0), ("h", 0), ("Fe2O3", 0), ("50", 0), ("Fe ", 0), ("", 0)
])
# fmt: on
def test_get_element_atomic_number(element_number):

    element_str, atomic_number = element_number
    assert (
        get_element_atomic_number(element_str) == atomic_number
    ), "Atomic number returned by the function is incorrect"


# fmt: off
@pytest.mark.parametrize("element_valid", [
    ("He", True), ("H", True), ("Fe", True),
    # Failing cases
    ("he", False), ("h", False), ("Fe2O3", False), ("50", False), ("Fe ", False), ("", False)
])
# fmt: on
def test_validate_element_str(element_valid):

    element_str, is_valid = element_valid
    assert validate_element_str(element_str) == is_valid, "Element validation is not successful"


# fmt: off
@pytest.mark.parametrize("formula, elements, n_atoms", [
    ("Fe2O3", ("Fe", "O"), (2, 3)),
    ("He", ("He",), (1,)),
    ("H2SO4", ("H", "S", "O"), (2, 1, 4)),
    ("C", ("C",), (1,)),
    ("C2H5OH", ("C", "H", "O"), (2, 6, 1))
])
# fmt: on
def test_parse_compound_formula1(formula, elements, n_atoms):
    data = parse_compound_formula(formula)
    assert len(data) == len(elements), "The number of parsed elements is incorrect"
    assert set(elements) == set(data.keys()), "The set of parsed elements is incorrect"
    assert (
        tuple(data[e]["nAtoms"] for e in elements) == n_atoms
    ), "The number of atoms in parsed data is determined incorrectly"


# fmt: off
@pytest.mark.parametrize("formula, element_mass_fraction", [
    ("Fe2O3", {"Fe": 0.6994364433312461, "O": 0.30056355666875395}),
    ("C", {"C": 1.0})
])
# fmt: on
def test_parse_compound_formula2(formula, element_mass_fraction):
    # Verify that mass fraction is found correctly
    data = parse_compound_formula(formula)
    assert len(data) == len(element_mass_fraction), "The number of elements in the parsed data is incorrect"
    for e in element_mass_fraction.keys():
        assert e in data, f"Element {e} is not found in parsed data"
        npt.assert_almost_equal(
            data[e]["massFraction"],
            element_mass_fraction[e],
            err_msg=f"Mass fraction for element {e} was evaluated incorrectly",
        )


# fmt: off
@pytest.mark.parametrize("formula", [
    "FE2O3", "fe2O3", "D", "Abc", ""
])
@pytest.mark.filterwarnings("ignore:")  # Ignore the warnings from XRayLib
# fmt: on
def test_parse_compound_formula_fail(formula):
    with pytest.raises(RuntimeError, match=f"Invalid chemical formula.*{formula}"):
        parse_compound_formula(formula)


# fmt: off
@pytest.mark.parametrize("formula, elements, n_atoms", [
    ("Fe2O3", ("Fe", "O"), (2, 3)),
    ("He", ("He",), (1,)),
    ("H2SO4", ("H", "S", "O"), (2, 1, 4)),
    ("C", ("C",), (1,))
])
# fmt: on
def test_split_compound_mass(formula, elements, n_atoms):

    mass_total = 10.0

    data = split_compound_mass(formula, mass_total)
    # Basic checks for proper parsing
    assert len(data) == len(elements), "The number of parsed elements is incorrect"
    assert set(elements) == set(data.keys()), "The set of parsed elements is incorrect"
    # Make sure that the sum of mass of each element equals to total mass
    npt.assert_almost_equal(
        np.sum(list(data.values())),
        mass_total,
        err_msg="The computed mass is not distributed properly among elements",
    )


# fmt: off
@pytest.mark.parametrize("formula", [
    "FE2O3", "fe2O3", "D", "Abc", ""
])
@pytest.mark.filterwarnings("ignore:")  # Ignore the warnings from XRayLib
# fmt: on
def test_split_compound_mass_fail(formula):
    with pytest.raises(RuntimeError, match=f"Invalid chemical formula.*{formula}"):
        split_compound_mass(formula, 10.0)


# --------------------------------------------------------------------------------------


def test_get_supported_eline_list():

    list_k = get_supported_eline_list(lines=("K",))
    list_l = get_supported_eline_list(lines=("L",))
    list_m = get_supported_eline_list(lines=("M",))

    list_kl = get_supported_eline_list(lines=("K", "L"))
    list_lm = get_supported_eline_list(lines=("L", "M"))

    list_klm = get_supported_eline_list(lines=("K", "L", "M"))
    list_default = get_supported_eline_list()

    # Check eline formatting
    for v in list_klm:
        assert re.search(r"[A-Z][a-z]?_[KLM]", v), f"Emission line name {v} is not properly formatted"

    assert (
        (len(list_k) > 0) and (len(list_l) > 0) and (len(list_m) > 0)
    ), "At least one of the lists for K, L or M lines is empty"
    assert list_klm == list_default, "The complete list of emission lines is incorrectly assembled"
    assert list_klm == list_k + list_l + list_m, "The complete list does not include all for K, L and M lines"
    assert list_kl == list_k + list_l, "The list for K and L lines is not equivalent to the sum of the lists"
    assert list_lm == list_l + list_m, "The list for K and L lines is not equivalent to the sum of the lists"


# fmt: off
@pytest.mark.parametrize("eline, success", [
    ("Fe_K", True),
    ("W_L", True),
    ("Ta_M", True),
    ("Ab_K", False),
    ("Fe_A", False),
    ("", False),
    (None, False)
])
# fmt: on
def test_check_if_eline_supported(eline, success):
    assert check_if_eline_supported(eline) == success, f"Emission line {eline} is indentified incorrectly"


# fmt: off
@pytest.mark.parametrize("eline, incident_energy, success", [
    ("Fe_K", 8.0, True),
    ("Fe_K", 6.0, False),
    ("W_L", 12.0, True),
    ("W_L", 8.0, False),
    ("Ta_M", 2.0, True),
    ("Ta_M", 1.0, False),

])
# fmt: on
def test_check_if_eline_is_activated(eline, incident_energy, success):
    assert (
        check_if_eline_is_activated(eline, incident_energy) == success
    ), f"Activation status for the emission line {eline} at {incident_energy} keV is {success}"


# fmt: off
@pytest.mark.parametrize("elements, incident_energy, elines", [
    (["Fe", "W", "Ta"], 12.0, ['Fe_K', 'W_L', 'W_M', 'Ta_L', 'Ta_M']),
    (["Fe", "W", "Ta"], 6.0, ['W_M', 'Ta_M']),
])
# fmt: on
def test_generate_eline_list1(elements, incident_energy, elines):
    r"""
    ``generate_eline_list``: search all lines
    """
    assert (
        generate_eline_list(elements, incident_energy=incident_energy) == elines
    ), "Emission line list is generated incorrectly"


# fmt: off
@pytest.mark.parametrize("elements, incident_energy, lines, elines", [
    (["Fe", "W", "Ta"], 12.0, ("K", "L", "M"), ['Fe_K', 'W_L', 'W_M', 'Ta_L', 'Ta_M']),
    (["Fe", "W", "Ta"], 12.0, ("K",), ['Fe_K']),
    (["Fe", "W", "Ta"], 12.0, ("L",), ['W_L', 'Ta_L']),
    (["Fe", "W", "Ta"], 12.0, ("M",), ['W_M', 'Ta_M']),
    (["Fe", "W", "Ta"], 12.0, ("K", "L"), ['Fe_K', 'W_L', 'Ta_L']),
    (["Fe", "W", "Ta"], 12.0, ("K", "M"), ['Fe_K', 'W_M', 'Ta_M']),
    (["Fe", "W", "Ta"], 12.0, ("L", "M"), ['W_L', 'W_M', 'Ta_L', 'Ta_M']),
])
# fmt: on
def test_generate_eline_list2(elements, incident_energy, lines, elines):
    r"""
    ``generate_eline_list``: explicitely select eline categories
    """
    assert (
        generate_eline_list(elements, incident_energy=incident_energy, lines=lines) == elines
    ), "Emission line list is generated incorrectly"


def test_generate_eline_list3():
    r"""
    ``generate_eline_list``: fails if emission line category is specified incorrectly
    """
    with pytest.raises(RuntimeError, match="Some of the selected emission lines are incorrect"):
        generate_eline_list(["Fe", "W", "Ta"], incident_energy=12.0, lines=["K", "Ka"])



import textwrap

from qtpy.QtWidgets import QHBoxLayout, QVBoxLayout, QDialog, QDialogButtonBox, QTextEdit, QLabel
from qtpy.QtCore import Qt

import logging

logger = logging.getLogger(__name__)


class DialogViewCalibStandard(QDialog):
    def __init__(self, parent=None, *, file_path="", calib_preview=""):

        super().__init__(parent)

        self.setWindowTitle("View Calibration Standard")

        self.setMinimumSize(300, 400)
        self.resize(700, 700)

        # Displayed data (must be set before the dialog is shown
        self.file_path = file_path
        self.calib_preview = calib_preview

        vbox = QVBoxLayout()
        hbox = QHBoxLayout()
        hbox.addWidget(QLabel("<b>Source file:</b> "), 0, Qt.AlignTop)
        file_path = textwrap.fill(self.file_path, width=80)
        hbox.addWidget(QLabel(file_path), 0, Qt.AlignTop)
        hbox.addStretch(1)
        vbox.addLayout(hbox)

        te = QTextEdit()
        te.setReadOnly(True)
        te.setText(self.calib_preview)
        vbox.addWidget(te)

        button_box = QDialogButtonBox(QDialogButtonBox.Close)
        button_box.accepted.connect(self.accept)
        button_box.rejected.connect(self.reject)
        vbox.addWidget(button_box)

        self.setLayout(vbox)
